# flake8: noqa

# This file is autogenerated by /metadata-ingestion/scripts/avro_codegen.py
# Do not modify manually!

# fmt: off
from .....schema_classes import AssertionInfoClass
from .....schema_classes import AssertionResultClass
from .....schema_classes import AssertionResultTypeClass
from .....schema_classes import AssertionRunEventClass
from .....schema_classes import AssertionRunStatusClass
from .....schema_classes import AssertionStdAggregationClass
from .....schema_classes import AssertionStdOperatorClass
from .....schema_classes import AssertionStdParameterClass
from .....schema_classes import AssertionStdParameterTypeClass
from .....schema_classes import AssertionStdParametersClass
from .....schema_classes import AssertionTypeClass
from .....schema_classes import BatchSpecClass
from .....schema_classes import DatasetAssertionInfoClass
from .....schema_classes import DatasetAssertionScopeClass


AssertionInfo = AssertionInfoClass
AssertionResult = AssertionResultClass
AssertionResultType = AssertionResultTypeClass
AssertionRunEvent = AssertionRunEventClass
AssertionRunStatus = AssertionRunStatusClass
AssertionStdAggregation = AssertionStdAggregationClass
AssertionStdOperator = AssertionStdOperatorClass
AssertionStdParameter = AssertionStdParameterClass
AssertionStdParameterType = AssertionStdParameterTypeClass
AssertionStdParameters = AssertionStdParametersClass
AssertionType = AssertionTypeClass
BatchSpec = BatchSpecClass
DatasetAssertionInfo = DatasetAssertionInfoClass
DatasetAssertionScope = DatasetAssertionScopeClass

# fmt: on
