{
  "type": "record",
  "name": "MLModelDeploymentProperties",
  "namespace": "com.linkedin.pegasus2avro.ml.metadata",
  "doc": "Properties associated with an ML Model Deployment",
  "fields": [
    {
      "name": "customProperties",
      "type": {
        "type": "map",
        "values": "string"
      },
      "doc": "Custom property bag.",
      "default": {},
      "Searchable": {
        "/*": {
          "queryByDefault": true
        }
      }
    },
    {
      "name": "externalUrl",
      "type": [
        "null",
        "string"
      ],
      "doc": "URL where the reference exist",
      "default": null,
      "java": {
        "class": "com.linkedin.pegasus2avro.common.url.Url",
        "coercerClass": "com.linkedin.pegasus2avro.common.url.UrlCoercer"
      }
    },
    {
      "name": "description",
      "type": [
        "null",
        "string"
      ],
      "doc": "Documentation of the MLModelDeployment",
      "default": null,
      "Searchable": {
        "fieldType": "TEXT",
        "hasValuesFieldName": "hasDescription"
      }
    },
    {
      "name": "createdAt",
      "type": [
        "null",
        "long"
      ],
      "doc": "Date when the MLModelDeployment was developed",
      "default": null
    },
    {
      "name": "version",
      "type": [
        "null",
        {
          "type": "record",
          "name": "VersionTag",
          "namespace": "com.linkedin.pegasus2avro.common",
          "doc": "A resource-defined string representing the resource state for the purpose of concurrency control",
          "fields": [
            {
              "name": "versionTag",
              "type": [
                "null",
                "string"
              ],
              "default": null
            }
          ]
        }
      ],
      "doc": "Version of the MLModelDeployment",
      "default": null
    },
    {
      "name": "status",
      "type": [
        "null",
        {
          "type": "enum",
          "name": "DeploymentStatus",
          "doc": "Model endpoint statuses",
          "symbols": [
            "OUT_OF_SERVICE",
            "CREATING",
            "UPDATING",
            "ROLLING_BACK",
            "IN_SERVICE",
            "DELETING",
            "FAILED",
            "UNKNOWN"
          ],
          "symbolDocs": {
            "CREATING": "Deployments being created.",
            "DELETING": "Deployments being deleted.",
            "FAILED": "Deployments with an error state.",
            "IN_SERVICE": "Deployments that are active.",
            "OUT_OF_SERVICE": "Deployments out of service.",
            "ROLLING_BACK": "Deployments being reverted to a previous version.",
            "UNKNOWN": "Deployments with unknown/unmappable state.",
            "UPDATING": "Deployments being updated."
          }
        }
      ],
      "doc": "Status of the deployment",
      "default": null
    }
  ],
  "Aspect": {
    "name": "mlModelDeploymentProperties"
  }
}